///////////////////////////////////
//Technik 2009
//Revolution Engine 0.5
///////////////////////////////////
//Materials source file

#include "REV.h"

MATERIAL::MATERIAL()
{
	flags = 0;
}

BMATERIAL::BMATERIAL(GXColor diffuse, GXColor specular, f32 specularity)
{
	type = MTT_BASIC;
	diffuseClr = diffuse;
	specularClr = specular;
	this->specularity = specularity;
	diffuseMap = NULL;
	specularMap = NULL;
	guMtxIdentity(diffMtx);
	guMtxIdentity(specMtx);
}

void BMATERIAL::setTev(u8 nUVs, u8 * clrChannels, u8 * texChannels)//Only shows shadows
{
	u8 nTex = 0;
	//Diffuse channel
	GX_SetChanCtrl(GX_COLOR0A0,GX_ENABLE, GX_SRC_VTX, GX_SRC_REG, GX_LIGHTNULL, GX_DF_CLAMP, GX_AF_NONE );
	GX_SetChanMatColor(GX_COLOR0A0, diffuseClr);
	//Specular channel
	GX_SetChanCtrl(GX_COLOR1A1,GX_ENABLE, GX_SRC_VTX, GX_SRC_REG, GX_LIGHTNULL, GX_DF_CLAMP, GX_AF_NONE );
	GX_SetChanMatColor(GX_COLOR1A1, specularClr);
	//Vertex descriptors
	GX_SetVtxDesc(GX_VA_CLR0, GX_DIRECT);//Diffuse lighting
	GX_SetVtxDesc(GX_VA_CLR1, GX_DIRECT);//Specular lighting
	*clrChannels = CC_DIFF + CC_SPEC;
	GX_SetNumChans(2);
	u8 currTev = GX_TEVSTAGE0;
	if(mainRoot->shadowsEnabled())
	{
		GX_SetTevOrder(currTev, GX_TEXCOORD1, GX_TEXMAP0, GX_COLOR0A0);
		GX_SetTevOp(currTev,GX_REPLACE);
		//Inputs
		//GX_SetTevColorIn(currTev, GX_CC_TEXC, GX_CC_KONST, GX_CC_ONE, GX_CC_ZERO);
		//GX_SetTevAlphaIn(currTev, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_TEXA);//Ignore alpha
		//Operations
		//GX_SetTevColorOp(currTev, GX_TEV_COMP_R8_EQ, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
		//GX_SetTevAlphaOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
		//GX_SetTevKColorSel(currTev,GX_TEV_KCSEL_K0);//Id
		GX_SetNumTexGens(1);
		GX_SetTexCoordGen(GX_TEXCOORD1, GX_TG_MTX3x4, GX_TG_POS, GX_TEXMTX0);
		currTev++;//One more tev stage
	}
	//TEVSTAGE 0
	if(diffuseMap && nUVs)//Diffuse mapping
	{
		GX_LoadTexObj(diffuseMap, GX_TEXMAP1);
		GX_SetTevColorIn(currTev, GX_CC_ZERO, GX_CC_RASC, GX_CC_TEXC, GX_CC_ZERO);
		GX_SetTevAlphaIn(currTev, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_TEXA);
		nTex++;
	}
	else
	{
		GX_SetTevColor(GX_TEVREG0,diffuseClr);//Diffuse color
		GX_SetTevColorIn(currTev, GX_CC_ZERO, GX_CC_ZERO, GX_CC_ZERO, GX_CC_RASC);
		GX_SetTevAlphaIn(currTev, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_RASA);
	}
	GX_SetTevOrder(currTev, GX_TEXCOORD0, GX_TEXMAP1, GX_COLOR0A0);
	GX_SetTevColorOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVREG2);
	GX_SetTevAlphaOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVREG2);
	currTev++;
	//TEVSTAGE 1
	if(specularMap && nUVs)
	{
		GX_LoadTexObj(specularMap, GX_TEXMAP2);
		GX_SetTevColorIn(currTev, GX_CC_ZERO, GX_CC_RASC, GX_CC_TEXC, GX_CC_ZERO);
		GX_SetTevAlphaIn(currTev, GX_CA_ZERO, GX_CA_RASA, GX_CA_TEXA, GX_CA_ZERO);
		GX_SetTevOrder(currTev, GX_TEXCOORD0, GX_TEXMAP2, GX_COLOR1A1);
		nTex++;
	}
	else
	{
		GX_SetTevColor(GX_TEVREG1,specularClr);
		GX_SetTevColorIn(currTev, GX_CC_ZERO, GX_CC_ZERO, GX_CC_ZERO, GX_CC_RASC);
		GX_SetTevAlphaIn(currTev, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_RASA);
		GX_SetTevOrder(currTev, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR1A1);
	}
	GX_SetTevColorOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevAlphaOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	//TEVSTAGE 2
	currTev++;
	GX_SetTevColorIn(currTev, GX_CC_ZERO, GX_CC_CPREV, GX_CC_APREV, GX_CC_C2);
	GX_SetTevAlphaIn(currTev, GX_CA_A2, GX_CA_ZERO, GX_CA_ZERO, GX_CA_APREV);
	GX_SetTevColorOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevAlphaOp(currTev, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
	GX_SetTevOrder(currTev, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0);
	currTev++;
	GX_SetNumTevStages(currTev);
	if(nTex)
	{
		GX_SetVtxDesc(GX_VA_TEX0, GX_INDEX16);
		*texChannels = 1;
	}
}

GXTexObj * loadPng(const char * filename)
{
	FILE * temp = fopen(filename, "rb");
	if(!temp)
		return NULL;
	else fclose(temp);
	GXTexObj * map = (GXTexObj*)memalign(32, sizeof(GXTexObj));
	PNGUPROP imgProp;
	IMGCTX ctx = PNGU_SelectImageFromDevice(filename);
	PNGU_GetImageProperties(ctx,&imgProp);
	void * data = memalign(32, imgProp.imgWidth * imgProp.imgHeight * 4);
	PNGU_DecodeTo4x4RGBA8 (ctx, imgProp.imgWidth, imgProp.imgHeight, data, 255);
	PNGU_ReleaseImageContext (ctx);
	DCFlushRange (data, imgProp.imgWidth * imgProp.imgHeight * 4);
	GX_InitTexObj (map, data, imgProp.imgWidth, imgProp.imgHeight, GX_TF_RGBA8, GX_CLAMP, GX_CLAMP, GX_FALSE);
	return map;//And return the texture
}
