///////////////////////////////////
//Technik 2009
//Revolution Engine 0.5
///////////////////////////////////
//Nodes source file
#include "REV.h"

NODE::NODE(u8 flags)
{
	this->flags = flags;
	type = NT_NODE;
	guMtxIdentity(base);
	child = NULL;
	if(mainRoot->rootNode)
	{
		next = mainRoot->rootNode->child;
		mainRoot->rootNode->child = this;
		parent = mainRoot->rootNode;
	}
	else
	{
		next = NULL;
		parent = NULL;
	}
	custom = NULL;
}

NODE::~NODE()
{
	releaseChildren();
	if(parent->child == this)
	{
		parent->child = next;
	}
	else
	{
		NODE * aux = parent->child;
		while(aux->next != this)
		{
			aux = aux->next;
		}
		aux->next = next;
	}
}

void NODE::attach(NODE * parent, u8 coord)
{
	deattach(coord);
	mainRoot->rootNode->child = next;
	next = parent->child;
	parent->child = this;
	this->parent = parent;
	if(coord)
	{
		Mtx m;
		parentMtx(m);
		guMtxInverse(m, m);
		guMtxConcat(m, base, base);
	}
}

void NODE::addChild(NODE * child, u8 coord)
{
	child->attach(this, coord);
}

void NODE::deattach(u8 coord)
{
	if(coord)
	{
		Mtx m;
		parentMtx(m);
		guMtxConcat(m, base, base);
	}
	if(parent->child == this)
	{
		parent->child = next;
	}
	else
	{
		NODE * aux = parent->child;
		while(aux->next != this)
		{
			aux = aux->next;
		}
		aux->next = next;
	}
	parent = mainRoot->rootNode;
	next = mainRoot->rootNode->child;
	mainRoot->rootNode->child = this;
}

void NODE::releaseChildren(u8 coord)
{
	while(child)
	{
		child->deattach(coord);
	}
}

void NODE::clearChildren()
{
	while(child)
	{
		child->clearChildren();
		delete child;
	}
}

void NODE::move(Vector move, u8 reference)
{
	Vector v;
	switch(reference)
	{
		case RT_SELF:
		{
			guVecMultiplySR(base, &move, &v);
			break;
		}
		case RT_PAR:
		{
			v = move;
			break;
		}
		case RT_ABS:
		{
			Mtx m;
			parentMtx(m);
			guVecMultiplySR(m, &move, &v);
			break;
		}
		default:
		{
			v = nullVector;
		}
	}
	guMtxTransApply(base, base, v.x, v.y, v.z);
}

void NODE::rotate(Vector axis, f32 angle, u8 reference)
{
	Vector pos = getPos(RT_PAR);
	guMtxTransApply(base, base, -pos.x, -pos.y, -pos.z);
	Mtx aux;
	switch(reference)
	{
		case RT_SELF:
		{
			guVecMultiply(base, &axis, &axis);
			break;
		}
		case RT_ABS://Absolute
		{
			parentMtx(aux);
			guVecMultiply(aux, &axis, &axis);
			break;
		}
	}
	guMtxRotAxisDeg(aux, &axis, angle);
	guMtxConcat(aux , base, base);
	guMtxTransApply(base, base, pos.x, pos.y, pos.z);
}

void NODE::setPos(Vector pos, u8 reference)
{
	Mtx aux;
	if(!reference)
	{
		parentMtx(aux);
		guMtxInverse(aux, aux);
		guVecMultiply(aux, &pos, &pos);
	}
	base[0][3] = pos.x;
	base[1][3] = pos.y;
	base[2][3] = pos.z;
}

Vector NODE::getPos(u8 reference)
{
	Vector v = {base[0][3], base[1][3], base[2][3]};
	if(!reference)
	{
		Mtx aux;
		parentMtx(aux);
		guVecMultiply(aux, &v, &v);
	}
	return v;
}

void NODE::parentMtx(Mtx m)
{
	guMtxIdentity(m);
	NODE * aux = parent;
	while(aux)
	{
		guMtxConcat(aux->base, m, m);
		aux = aux->parent;
	}
}

void NODE::absMtx(Mtx m)
{
	parentMtx(m);
	guMtxConcat(m, base, m);
}

void NODE::srMtx(Mtx m)
{
	absMtx(m);
	m[0][3] = 0.0f;
	m[1][3] = 0.0f;
	m[2][3] = 0.0f;
}

void NODE::lookAt(Vector look)
{
	base[0][1] = look.x;
	base[1][1] = look.y;
	base[2][1] = look.z;
	//
	base[0][0] = look.y;
	base[1][0] = -look.x;
	base[2][0] = 0;
	//
	base[0][2] = -look.x*look.z;
	base[1][2] = -look.y*look.z;
	base[2][2] = look.x*look.x;
}

N2D::N2D(u8 flags)
{
	this->flags = flags;
	type = NT_NODE;
	guMtxIdentity(base);
	this->angle = 0;
	//now Queuing
	child = NULL;
	if(mainRoot->rootN2D)
	{
		next = mainRoot->rootN2D->child;
		mainRoot->rootN2D->child = this;
		parent = mainRoot->rootN2D;
	}
	else
	{
		next = NULL;
		parent = NULL;
	}
	custom = NULL;
}

N2D::~N2D()
{
	releaseChildren();
	if(parent->child == this)
	{
		parent->child = next;
	}
	else
	{
		N2D * aux = parent->child;
		while(aux->next != this)
		{
			aux = aux->next;
		}
		aux->next = next;
	}
}

void N2D::attach(N2D * parent, u8 coord)
{
	deattach(coord);
	mainRoot->rootN2D->child = next;
	next = parent->child;
	parent->child = this;
	this->parent = parent;
	if(coord == RT_ABS)
	{
		Mtx m;
		parent->absMtx(m);
		guMtxInverse(m, m);
		guMtxConcat(m, base, base);
	}
}

void N2D::addChild(N2D * child, u8 coord)
{
	child->attach(this, coord);
}

void N2D::deattach(u8 coord)
{
	if(coord == RT_ABS)
	{
		Mtx m;
		parentMtx(m);
		guMtxConcat(m, base, base);
	}
	if(parent->child == this)
	{
		parent->child = next;
	}
	else
	{
		N2D * aux = parent->child;
		while(aux->next != this)
		{
			aux = aux->next;
		}
		aux->next = next;
	}
	parent = mainRoot->rootN2D;
	next = mainRoot->rootN2D->child;
	mainRoot->rootN2D->child = this;
}

void N2D::releaseChildren(u8 coord)
{
	while(child)
	{
		child->deattach(coord);
	}
}

void N2D::clearChildren()
{
	while(child)
	{
		child->clearChildren();
		delete child;
	}
}

void N2D::move(Vector move, u8 reference)//buggy?
{
	Vector v;
	switch(reference)
	{
		case RT_SELF:
		{
			guVecMultiplySR(base, &move, &v);
			break;
		}
		case RT_PAR:
		{
			v = move;
			break;
		}
		case RT_ABS:
		{
			Mtx m;
			parentMtx(m);
			guVecMultiplySR(m, &move, &v);
			break;
		}
		default:
		{
			v = nullVector;
		}
	}
	guMtxTransApply(base, base, v.x, v.y, v.z);
}

void N2D::rotate(f32 angle)//buggy?
{
	Mtx rot;
	guMtxRotDeg(rot, 'z', angle);
	f32 x0 = base[0][3];
	f32 y0 = base[1][3];
	base[0][3] = base[1][3] = 0.0f;
	guMtxConcat(rot, base, base);
	base[0][3] = x0;
	base[1][3] = y0;
	this->angle += angle;
}

void N2D::setPos(Vector pos, u8 reference)
{
	Mtx aux;
	if(reference == RT_ABS)
	{
		parentMtx(aux);
		guMtxInverse(aux, aux);
		guVecMultiply(aux, &pos, &pos);
	}
	base[0][3] = pos.x;
	base[1][3] = pos.y;
	base[2][3] = pos.z;
}

Vector N2D::getPos(u8 coord)//
{
	Mtx aux;
	Vector v = {base[0][3], base[1][3], base[2][3]};
	if(coord == RT_ABS)
	{
		parentMtx(aux);
		guVecMultiply(aux, &v, &v);
	}
	return v;
}

void N2D::setAngle(f32 angle, u8 reference)
{
	f32 x= base[0][3];
	f32 y= base[1][3];
	f32 layer = base[2][3];
	guMtxRotDeg(base, 'z', angle);
	if(!reference)
	{
		Mtx m;
		parentMtx(m);
		guMtxInverse(m, m);
		guMtxConcat(m, base, base);
	}
	base[0][3] = x;
	base[1][3] = y;
	base[2][3] = layer;
	this->angle = angle;
}

f32 N2D::getAngle(u8 coord)
{
	if(coord != RT_ABS)
		return this->angle;
	N2D * aux = this;
	f32 ang = 0;
	while(aux->parent)
	{
		ang += aux->angle;
		aux = aux->parent;
	}
	return ang;
}

void N2D::parentMtx(Mtx m)
{
	guMtxIdentity(m);
	N2D * aux = parent;
	while(aux)
	{
		guMtxConcat(aux->base, m, m);
		aux = aux->parent;
	}
}

void N2D::absMtx(Mtx m)
{
	parentMtx(m);
	guMtxConcat(m, base, m);
}

void N2D::srMtx(Mtx m)
{
	absMtx(m);
	m[0][3] = 0.0f;
	m[1][3] = 0.0f;
	m[2][3] = 0.0f;
}
