///////////////////////////////////
//Technik 2009
//Revolution Engine 0.5
///////////////////////////////////
//Viewports source file
#include "REV.h"

#define screenHeightMM	30.0f

VIEWPORT::VIEWPORT(Vector pos, f32 sx, f32 sy, CAMERA * cam, u8 flags)
{
	this->flags = flags;
	this->sx = sx;
	this->sy = sy;
	//Allocate memory for the texture
	buffer = memalign(32, GX_GetTexBufferSize(sx, sy, GX_TF_RGBA8, GX_FALSE,0));
	img = (GXTexObj*)memalign(32, sizeof(GXTexObj));
	camera = cam;
	nextVP = mainRoot->views;
	mainRoot->views = this;
	setPos(pos);
}

VIEWPORT::~VIEWPORT()
{
	VIEWPORT * v = mainRoot->views;
	if(v == this) mainRoot->views = nextVP;
	else
	{
		while(v->nextVP != this)
			v = v->nextVP;
		v->nextVP = nextVP;
	}
	free(img);
	free(buffer);
}

void VIEWPORT::setSize(f32 sx, f32 sy)
{
	if(this->sx * this-> sy != sx*sy)
	{
		free(buffer);
		buffer = memalign(32, GX_GetTexBufferSize(sx, sy, GX_TF_RGBA8, GX_FALSE,0));
	}
	this->sx = sx;
	this->sy = sy;
}

void VIEWPORT::setForRender(Mtx44 m)
{
	if(camera->type == NT_CAM)
		guPerspective(m, camera->fov, sx/sy, camera->ncd, camera->fcd);
	if(camera->type == NT_HTCAM)
	{
		Vector v = camera->getPos(); 
		guFrustum(m, -screenHeightMM*camera->ncd*(0.5 * screenHeightMM* sx/sy + v.x)/v.y
					, -screenHeightMM*camera->ncd*(-0.5f * screenHeightMM*sx/sy + v.x)/v.y
					, screenHeightMM*camera->ncd*(0.5f + v.z)/v.y
					, screenHeightMM*camera->ncd*(0.5f + v.z)/v.y, camera->ncd, camera->fcd);
	}
	GX_LoadProjectionMtx(m, GX_PERSPECTIVE);
	GX_SetViewport(0,0,sx,sy,0.0f,1.0f);
	GX_SetScissor(0,0,sx,sy);
}

void VIEWPORT::copyTexture()
{
	GX_SetCopyFilter(GX_FALSE, NULL, GX_FALSE, NULL);
	GX_SetTexCopySrc(0, 0, sx, sy);
	GX_SetTexCopyDst(sx, sy, GX_TF_RGBA8, GX_FALSE);
	GX_CopyTex(buffer, GX_TRUE);
	GX_PixModeSync();
	//
	GX_InitTexObj (img, buffer, sx, sy, GX_TF_RGBA8, GX_CLAMP, GX_CLAMP, GX_FALSE);
}
